\ DIR with date/time stamp
\ Source text for DX-Forth
\ Pavel Zampach, 2020
\ *

forth definitions decimal

variable #ENTRY                         \ number of dir entries
$80 constant DEFDMA                     \ default dma
$5C constant DEFFCB                     \ default fcb
28  constant ENTRYSIZE                  \ size of dir item
11  constant FNL                        \ filename length

\ Dir entry structure (28 bytes)
\ 12 bytes filename
\ 15 bytes ZSDOS file stamp
\  2 bytes number of file sectors

create WFCB                             \ work FCB
  36 allot

\ Index of entry array
: ENTRY()   ( n -- addr )
  entrysize * here + 256 +
;

\ Print complete filename
\ addr = address of entry
: .FILENAME    ( addr -- )
  dup 8 type                            \ filename
  [char] . emit                         \ dot
  8 + 3 type                            \ filetype
;

\ Print BCD
: .BCD        ( c -- )
  dup 4 rshift [char] 0 + emit
  $0f and [char] 0 + emit
;

: .DATE       ( addr -- )
  $10B c@ if
    \ American date
    dup 1+ c@ .bcd [char] / emit
    dup 2+ c@ .bcd [char] / emit
  else
    \ European date
    dup 2+ c@ .bcd [char] . emit
    dup 1+ c@ .bcd [char] . emit
  then
  c@ .bcd
;

\ Print file time, addr = address of datetime field
: .TIME       ( addr -- )
  dup 3 + c@ .bcd [char] : emit
  4 + c@ .bcd
;

\ Print file date and time, addr = address of entry
: .DATETIME   ( addr -- )
  21 + dup                              \ address of modify field
  .date space .time
;

\ Print file size, addr = address of entry
: .SIZE       ( addr -- )
  26 + @
  8 /mod swap if 1+ then
  3 .r [char] k emit
;

\ Dir entry (0-3) -> addr of dir entry
: DIR>ADR      ( direntry -- addr )
  32 * defdma +
;

\ Read number of file sectors
: READ-SEC    ( addr -- )
  wfcb 35 bdos drop
  wfcb 33 + @
  swap 26 + !
;

\ Read file stamp
: READ-DT     ( addr -- )
  fnl + dup                             \ address of file stamp
  15 $11 fill                           \ fill file stamp area by "1"
  setdma                                \ and set DMA to entry file stamp
  wfcb 102 bdos drop                    \ ZSDOS function 102 for file stamp reading
  defdma setdma                         \ restore DMA
;

\ Read filename
: READ-FN     ( addr -- )
  wfcb 1+ swap fnl cmove
;

: READ-LOOP   ( -- )
  #entry off
  defdma setdma
  deffcb 17 bdos                        \ first search
  begin
    dup 4 u<                            \ found?
  while
    wfcb 36 erase                       \ prepare work FCB
    dir>adr wfcb 12 cmove               \ fill filename
    deffcb c@ wfcb c!                   \ fill drive
    #entry @ entry() dup dup
    read-fn read-dt read-sec
    1 #entry +!
    0 18 bdos                           \ next search
  repeat   drop
;

\ Pause after [u1] items
: ?MORE        ( u1 u2 -- )
  1+ swap mod 0= if
    ." [more]" key
    3 = if rstusr quit then             \ ^C ends program
    13 emit                             \ to begin of line
  then
;

: .ENTRY      ( n -- )
  dup #entry @ < if
    entry() dup dup
    .filename space
    .size space
    .datetime
  then
;

: .SEPAR      ( -- )
  space space ." |  "
;

\ Vertical printout
: WR-LOOP-V   ( -- )
  #entry @ dup 2/ swap 1 and + dup 0 do
    i .entry .separ
    dup i + .entry cr
    20 i ?more
  loop drop
;

\ Horizontal printout
: WR-LOOP-H   ( -- )
  #entry @ dup 0 do
    i .entry
    i 1 and if cr else .separ then
    40 i ?more
  loop
  1 and if cr then
;

\ Comparing words: addr1 and addr2 are addresses od dir entry
\ flag is true if entry1 < entry2

\ Compare filename + filetype
: COMP-N      ( addr1 addr2 -- flag )
  fnl swap over compare 0<
;

\ Compare filetype + filename
: COMP-T      ( addr1 addr2 -- flag )
  2dup 8 + swap 8 + swap 3 swap over
  compare ?dup if
    nip nip 0<
  else
    comp-n
  then
;

\ Compare file date
: COMP-D      ( addr1 addr2 -- flag )
  21 + swap 21 + swap 5 swap over
  compare 0<
;

\ Compare size
: COMP-S      ( addr1 addr2 -- flag )
  26 + @ swap 26 + @ u>
;

defer COMP
  ' comp-n is comp
defer ORD
  ' noop is ord

variable K
create X
  entrysize allot

\ Move one entry item
: EMOVE       ( addr1 addr2 -- )
  entrysize cmove
;

\ Sort entry array by filename
: SORT        ( -- )
  #entry @ 1- 0 ?do                     \ for i:=1 to n-1
    i k !                               \ k:=i
    i entry() x emove                   \ x:=a[i]
    #entry @ i 1+ do                    \ for j:=i+1 to n
      i entry() x comp ord if           \ if a[j].key < x.key
        i k !                           \ k:=j
        i entry() x emove               \ x:=a[j]
      then
    loop
    i entry() k @ entry() emove         \ a[k]:=a[i]
    x i entry() emove                   \ a[i]:=x
  loop
;

: FLAGS       ( addr u -- )
  2dup upper
  2dup [char] T scan nip if ['] comp-t is comp then
  2dup [char] D scan nip if ['] comp-d is comp then
  2dup [char] S scan nip if ['] comp-s is comp then
  [char] - scan nip if ['] 0= is ord then
;

\ Process command line
: CLINE        ( -- )
  cmdtail [char] / scan over >r
  flags
  cmdtail drop r> over - -trailing
  2dup [char] . scan nip 0= if          \ is filename entered?
    s" *.*" 2swap +string               \ no -> append "*.*"
  then
  deffcb setfcb setusr                  \ init system
;

: MAIN         ( -- )
  cline
  cr ." QDir 1.0" cr
  read-loop
  #entry @ if
    sort
    $10c c@ if
      wr-loop-h                         \ horizontal prinout
    else
      wr-loop-v                         \ vertical printout
    then
    #entry ? 
  else
    ." No "
  then
  ." files"
  rstusr
;

turnkey main qdir
